/*
===============================================================================
 Name        : main.c
 Author      : $(author)
 Version     :
 Copyright   : $(copyright)
 Description : main definition
===============================================================================
*/

#ifdef __USE_CMSIS
#include "LPC17xx.h"
#endif

#include <cr_section_macros.h>
#include "stdlib.h"
#include "string.h"
#include "ssp.h"
#include "i2s.h"
#include "timer.h"
#include "i2c.h"
#include "type.h"
// libmad
#include "libmad/config.h"
#include "libmad/decoder.h"
// FatFS
#include "integer.h"
#include "diskio.h"
#include "ffconf.h"
#include "ff.h"
// 自己／デバッグ
#include "mp3_12.h"
//#include "uart.h"
#include "glcd.h"

#define SAMPLE_FREQ 			44100	// サンプリング周波数の初期値

#define INPUT_BUF_SIZE			(512*4)		// madlibのinput関数内バッファのサイズ 	2100

//#define TIMER0_INTERVAL			((SystemCoreClock/(4*1000))*50)		// 50ms (PCLK=CCLK/4)	// 音飛びする
//#define TIMER0_INTERVAL			((SystemCoreClock/(4*1000))*1)		// 1ms (PCLK=CCLK/4)
#define TIMER0_INTERVAL			((SystemCoreClock/(4*1000*10)))			// 0.1ms (PCLK=CCLK/4)
#define TIMER1_INTERVAL			((SystemCoreClock/(4*5)))				// 200ms=1/5s (PCLK=CCLK/4)
#define TIMER1_1MIN				(5*60)		// 1分
#define TIMER2_INTERVAL			(SystemCoreClock/4)						// 1s (PCLK=CCLK/4)
#define	TIMER2_PRESCALE			20

// ボタン操作
#define SW_PRESSED				0x01
#define SWPLAY_PRESSED          0x01
#define SWOPTION_PRESSED     	0x02
#define SWNEXT_PRESSED          0x03
#define SWDESELECT_PRESSED     	0x04
#define SWPREV_PRESSED          0x05
#define SWUP_PRESSED            0x06
#define SWSELECT_PRESSED      	0x07
#define SWDOWN_PRESSED          0x08
#define SWPOFF_PRESSED			0x09

#define	SWSHIFT_MASK			0x10

#define	SWUPSHIFT_PRESSED		(SWUP_PRESSED | SWSHIFT_MASK)
#define	SWDOWNSHIFT_PRESSED		(SWDOWN_PRESSED | SWSHIFT_MASK)
#define	SWNEXTSHIFT_PRESSED		(SWNEXT_PRESSED | SWSHIFT_MASK)
#define	SWPREVSHIFT_PRESSED		(SWPREV_PRESSED | SWSHIFT_MASK)

#define POWER_ON_BIT			31		// P1.31
#define SPEAKER_MUTE_BIT		26		// P0.26
#define MCP23017_RESET_PORT		 6		// P0.6
#define POWER_LED_BIT			24		// P0.24

#define	LCD_IMAGE_BUFFER_SIZE	(LCD_BYTE_HEIGHT * LCD_WIDTH)
#define MAX_SELDIR_ITEM			10		// DIR選択画面で一度に表示されるDIR数の最大数
#define MAX_SELMP3_ITEM         5       // TRACK選択画面で一度に表示される数
#define LCD_STRING_LINES		5		// LCDに表示する文字列行数(固定値)
#define LCD_ITEMS_MODE_5		5		// LCDに表示する項目数モード
#define LCD_ITEMS_MODE_10		10

// ディレクトリリストファイル DIRLISTFILE parameters
#define DIRLISTFILE             "/DIRLST10.DAT"
#define RECLEN_DIRLIST           128UL
#define MAXLEN_FILEPATH          60UL
#define MAXLEN_DIRPATH           47UL
#define DL_FULLDIR               0
#define DL_LENFULLDIR            61
#define DL_DIRNAME               62
#define DL_RECNO_SUBDIR_START    75
#define DL_NUM_SUBDIR            77
#define DL_RECNO_PARENTDIR       78
#define DL_RECNO_GRANDPARENTDIR  80
#define DL_RECNO_NEXTRANDOMDIR   82
#define DL_NUM_SNDFILE           84
#define DL_DISPLAYNAME           85
#define DL_WIDTH_DIRNMI          118
#define DL_SCROLL                120
#define	DL_DOTLINEHEIGHT		 12

#define	RECLEN_DIRNAMEIMG_THIRD  128UL			// LCD表示イメージの３ぶんの１のバイト数
#define RECLEN_DIRNAMEIMG        (3*RECLEN_DIRNAMEIMG_THIRD)     // LCD表示イメージの全バイト数
#define RECLEN_DIRLIST_FULL      (RECLEN_DIRLIST + RECLEN_DIRNAMEIMG)    // 全レコード長

// MP3ファイルリスト SNDFILELISTFILE parameters
#define SNDFILELISTFILE         "SNDLST10.DAT"
#define RECLEN_MP3LIST           64UL
#define ML_FILENAME              0
#define ML_TRACKNO               13
#define ML_TITLENAME             16
#define ML_WIDTH_SNDNMI          49
#define ML_SCROLL                51

#define	RECLEN_TITMENAMEIMG_THIRD 	128UL			// OLED表示イメージの上または下半分
#define	RECLEN_TITMENAMEIMG     	(3*RECLEN_TITMENAMEIMG_THIRD)
#define RECLEN_MP3LIST_FULL     	(RECLEN_MP3LIST + RECLEN_TITMENAMEIMG)

// ボリュームファイル VOLUMEFILE parameters
#define VOLUMEFILE              "VOLUME12.DAT"
#define RECLEN_VOLUME           5UL		// 2UL
#define VF_VOLUME               0		// Speaker On時
#define VF_VOLUME_2             2		// Speaker Off時
#define VF_COUNT                4		// 1

// 履歴ファイル HISTFILE parameters
#define HISTFILE                "/HIST04.DAT"
#define RECLEN_HISTFILE         10UL
#define HS_LASTDIR              0
#define HS_LASTTRACK            2
#define HS_LASTBYTE             4
#define HS_NEXTRANDOMDIR        8

// メニューイメージファイル
#define MENUDATFILE				"/MNUDAT12.DAT"
#define	MAINMENU_SELDIR			0
#define	MAINMENU_SELINDEX		1
#define MAINMENU_RANDOM			2
#define MAINMENU_POFF			3
#define NUM_MAINMENU			3
#define OPTIONMENU_1			0
#define OPTIONMENU_2			1
#define OPTIONMENU_3			2
#define NUM_OPTIONMENU			3

#define COM_KND_MAINMENU		0x01	// メインメニュー
#define COM_KND_OPTIONMENU		0x05	// オプションメニュー

// インデックスファイル
#define INDEXFILE             	"/INDEX10.DAT"
#define IL_WIDTH_SELINDEX_ITEM	12				// １項目あたりのドット幅
#define	SELINDEX_LINES			4
#define NUM_SELINDEX_ITEM		(SELINDEX_LINES * 10)		// インデックス選択画面で一度に表示されるインデックス数(横１０文字)

// FatFs
static FATFS        *pFatfs;
static FIL 			*pfilMP3;

// DAC/madlib
typedef unsigned short		dacout_s;
dacout_s dacbuf[1152*4];
dacout_s *dac_s, *dac_e;
dacout_s *dac_s2, *dac_e2;
dacout_s *buftab[2];
int 	dacflg;
void	*pDecoderSync;		// decoder->sync 退避 (decoder.c)
void	*pStreamMainData;	// stream->main_data退避 (layer3.c)
void	*pFrameOverlap;		// frame->overlap退避 (layer3.c)
//volatile int DecodeSkipflg;		// 0:No skip 1:Skip in 2:Skip out
//unsigned long SkipCounter;
unsigned int samplerate;	// サンプルレートの退避
enum mad_flow (*output_func_backup)(void *, struct mad_header const *, struct mad_pcm *);	// output_funcの一時的退避

// イベント/制御
static volatile int	PinEvent;
static uint8_t      PonRestart;     // execute Power on Restart
static volatile uint16_t	CHATTERING_TMR;
static volatile uint16_t	AutoPoffTimer;
static uint8_t    	LcdImageBuff[LCD_IMAGE_BUFFER_SIZE]; 	// LCDフル画面イメージデータ
static uint8_t      PlayDirHist[RECLEN_HISTFILE];			// 履歴ファイルレコード
static uint8_t		SpeakerOn;		// 1=ON, 0=OFF
static uint8_t		ChannelMode;	// 1=STEREO, 0=MONO
static uint8_t		NowPlaying;

typedef enum {PM_RANDOM_DIR, PM_RANDOM_TRACK, PM_SELECT} playmode_t;
static playmode_t   PlayMode;

typedef enum {PS_PLAY, PS_SELECT} playorselect_t;

// 音量
volatile static mad_fixed_t		VolumeFactor;
volatile static uint16_t		Volume;
static uint16_t		VolumeTab[] = {8,10,13,16,20,25,32,40,50,63,
									79,100,126,158,199,251,316,398,501,630,
									793,999,1258,1583,1994,2510,3160,3978,5009,6306};	// MAX_VOLUME=10000
#define	MIN_VOLUMETAB_IDX		0
#define MAX_VOLUMETAB_IDX		((sizeof(VolumeTab)/sizeof(VolumeTab[0])) - 1)
#define	DEF_VOLUMETAB_IDX		15		// default
#define MIN_VOLUME              0		// Volume最小値
#define MAX_VOLUME              10000	// Volume最大値

// 関数宣言
static int decode();
int PauseAndSelectItem();
void SysTick_Off();
void SysTick_On();
void SystemInit (void);
void SystemClockUpdate (void);
void SystemCoreClockUpdate (void);
void InitApp();
void InitGlcdFull();
void PowerOff();
int ReadMainOrOptionMenuImage(FIL *Fil, uint8_t menukind);
int SelectDirectory(uint8_t *RecDirList, FIL *Fil);
FRESULT f_read_DIRLISTFILE (FIL* fp, void* buff);
int SelectIndex(uint8_t *RecDirList, FIL *Fil);
void SetRecDirList2MenuImage(uint8_t *RecDirList, uint8_t no, uint8_t itemsperpage);
void SetRecDirList2IndexMenuImage(uint8_t *RecDirList, uint8_t no);
void SetMenuItemSelected(uint8_t selecteditem, uint8_t itemsperpage);
void DispLcdFullImage();
uint8_t GetUsePage(uint8_t item, uint8_t itemsperpage);
void GetUseXaxis(uint8_t x[], uint8_t item, uint8_t itemsperpage);
void DispMenuItemsUpDown(uint8_t item1, uint8_t item2, uint8_t itemsperpage);
void DispMenuItemsLeftRight(uint8_t item1, uint8_t item2, uint8_t itemsperpage);
int PlayOrSelectMp3InSubdirectory(uint8_t *RecDirList, uint16_t dirno, uint16_t numsndfile, uint8_t start_trackno, playorselect_t playorselect);
int SelectOptionMenu(FIL *Fil);
int PlayRandom(uint8_t *RecDirList, FIL *Fil);
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno);
void SetImageAtPlaying(FIL *Fil, uint8_t *RecMp3List, uint8_t trackno, uint16_t numsndfile, uint8_t NumSelItem);
int ReadHistory(FIL *Fil);
void SaveHistory(FIL *Fil);
void ToggleSpeakerOnOff();
void ToggleChannelModeOnOff();


#if defined(MP3_12)
void __main(void) {
#else
int main(void) {
#endif

    FRESULT     res;
    int			ret;
    int         return_to_mainmenu;
    uint8_t     selectedmenu;
    uint8_t     selectedmenuBk;
    FIL 		Fil;
    uint8_t     RecDirList[RECLEN_DIRLIST_FULL];
    uint8_t     readdlflg;      			// メニューファイルを読む必要があるか？
    uint8_t		NumSelItem;					// １ページに表示する最大項目数

    SystemInit();
    //SystemClockUpdate();
    SystemCoreClockUpdate();

    InitApp();
    NumSelItem = LCD_ITEMS_MODE_5;
	return_to_mainmenu = 1;
    readdlflg = TRUE;
    CHATTERING_TMR = 0;
    AutoPoffTimer = 0;
    NowPlaying = 0;
    selectedmenu = 0;
    selectedmenuBk = 0;

	res = f_mount(pFatfs, _T("0:"), 0);
	if (res == FR_OK){
        ReadHistory(&Fil);

	    LPC_GPIO0->FIOSET |= (0x01 << SPEAKER_MUTE_BIT);	// P0.26 (set high)
	    SpeakerOn = 1;

		do {
        	//disp_akiduki_sample();
			//disp_akiduki_sample2();
            if (readdlflg == TRUE){
            	ReadMainOrOptionMenuImage(&Fil, COM_KND_MAINMENU);
                SetMenuItemSelected(selectedmenu, NumSelItem);
            	DispLcdFullImage();
            	readdlflg = FALSE;
            }

			ret = PauseAndSelectItem();
			switch (ret){
			case SWDOWN_PRESSED:
				selectedmenuBk = selectedmenu;
				selectedmenu++;
				if (selectedmenu == NUM_MAINMENU){
					selectedmenu = 0;
				}
				if (selectedmenu != selectedmenuBk){
		            SetMenuItemSelected(selectedmenuBk, NumSelItem);	// 選択状態を戻す
		            SetMenuItemSelected(selectedmenu, NumSelItem);
					DispMenuItemsUpDown(selectedmenu, selectedmenuBk, NumSelItem);
					readdlflg = FALSE;
				}
				break;
			case SWUP_PRESSED:
				selectedmenuBk = selectedmenu;
				if (selectedmenu > 0){
					selectedmenu--;
				} else {
					selectedmenu = NUM_MAINMENU - 1;
				}
				if (selectedmenu != selectedmenuBk){
		            SetMenuItemSelected(selectedmenuBk, NumSelItem);	// 選択状態を戻す
		            SetMenuItemSelected(selectedmenu, NumSelItem);
					DispMenuItemsUpDown(selectedmenu, selectedmenuBk, NumSelItem);
					readdlflg = FALSE;
				}
				break;
			case SWSELECT_PRESSED:
					switch (selectedmenu){
					case MAINMENU_SELDIR:			// ディレクトリ選択
   						PlayMode = PM_SELECT;
   						ret = SelectDirectory(RecDirList, &Fil);
   						if (ret != FR_OK && ret != GOTO_MAINMENU){
   							return_to_mainmenu = FALSE;
   						}
   						readdlflg = TRUE;
   						break;
   					case MAINMENU_SELINDEX:			// インデックスメニュー
   						PlayMode = PM_SELECT;
   						ret = SelectIndex(RecDirList, &Fil);
   						if (ret != FR_OK && ret != GOTO_MAINMENU){
   							return_to_mainmenu = FALSE;
   						}
   						readdlflg = TRUE;
   						break;
   					case MAINMENU_RANDOM:			// ランダム再生
   						PlayMode = PM_RANDOM_DIR;
   						PonRestart = 1;
   						ret = PlayRandom(RecDirList, &Fil);
   						if (ret != FR_OK && ret != GOTO_MAINMENU){
   							return_to_mainmenu = FALSE;
   						}
   						readdlflg = TRUE;
   						break;
   					default:
   						break;
   					}
				break;
			case SWOPTION_PRESSED:
    			ret = SelectOptionMenu(&Fil);
    			readdlflg = TRUE;
    			if (ret == OPTION_1){
    				InitGlcdFull();
    			} else if (ret == OPTION_2){
    				ToggleSpeakerOnOff();
    			} else if (ret == OPTION_3){
    				ToggleChannelModeOnOff();
    			} else if (ret == POWEROFF){
    				return_to_mainmenu = 0;
    			}
				break;
			case SWPOFF_PRESSED:
				return_to_mainmenu = 0;
				break;
			}
		} while (return_to_mainmenu);
	}

	PowerOff();
    //return 0 ;
}

/*
 * 各種初期化(IO/Timer/I2S/SPI/I2C/UART/GCLD/メモリ領域)
 */
void InitApp(){
    LPC_GPIO0->FIODIR |= (0x01 << SPEAKER_MUTE_BIT);// P0.26 (output)
	LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_MUTE_BIT);// P0.26 (set low)
    LPC_GPIO1->FIODIR |= (0x01 << POWER_ON_BIT);// P1.31 (output)
    LPC_GPIO0->FIODIR |= (0x01 << POWER_LED_BIT);// P0.24 (output)
    LPC_GPIO0->FIODIR |= ((0x01 << 23) | (0x01 << 25));// open pin

    LPC_GPIO0->FIODIR |= (0x01 << MCP23017_RESET_PORT);			// P0.6 (output/MCP23017 reset)
    LPC_GPIO0->FIOSET |= (0x01 << MCP23017_RESET_PORT);			// P0.6 (set high)

	LPC_PINCON->PINMODE3 &= ~(0x03 << 30);		// P1.31 (pull-up enabled)

	//LPC_SC->PCON |= 0x18;		// BOGD = 1, BORD=1

	init_timer( 0, TIMER0_INTERVAL );
	init_timer( 1, TIMER1_INTERVAL );
	//init_timer2( TIMER2_INTERVAL, TIMER2_PRESCALE);
	SSP0Init();		// ssp.c
	i2s_init();		// i2s.c
	enable_timer( 0 );
	enable_timer( 1 );

	LPC_GPIO0->FIOCLR |= (0x01 << MCP23017_RESET_PORT);		// P0.6 (MCP23017 reset low)
	I2C1Init();

//	UARTInit(1, 115200);
	//DebugWritelnV1("TIMER0_INTERVAL", TIMER0_INTERVAL);

    LPC_GPIO1->FIODIR |= ((0x01 << 18) | (0x01 << 20) | (0x01 << 21) | (0x01 << 23));		// on board led
    //LPC_GPIO0->FIODIR = ((0x01 << 29) | (0x01 << 30));	// P0.29/P0.30としては使えない

    // power swがリリースされるのを待つ
    while ((LPC_GPIO1->FIOPIN3 & (1 << 6)) == 0){	// P1.30

    }
    LPC_GPIO1->FIOSET |= (0x01 << POWER_ON_BIT);		// P1.31 (set high)

    LPC_GPIO0->FIOSET |= (0x01 << MCP23017_RESET_PORT);	// P0.6 (set high)
    mcp23017_init();
    glcd_init();

	pFatfs = (FATFS *)mad_malloc(sizeof(FATFS));
	pfilMP3 = (FIL *)mad_malloc(sizeof(FIL));
	pDecoderSync = 0;
	pStreamMainData = 0;
	pFrameOverlap = 0;

//	pStrmBuff = (unsigned char *)mad_malloc(INPUT_BUF_SIZE);
	dac_s = dac_e = dacbuf;
	dac_s2 = dac_e2 = dacbuf;
	buftab[0] = dacbuf;
	buftab[1] = &dacbuf[1152*2];
	dacflg = 0;

	//VolumeFactor = MAD_F_ONE;		//mad_f_tofixed((long)options.volume/100.0);
	Volume = VolumeTab[DEF_VOLUMETAB_IDX];
	VolumeFactor = mad_f_tofixed((long)Volume/(float)MAX_VOLUME);
	//DecodeSkipflg = DECODE_SKIP_OUT;

	samplerate = SAMPLE_FREQ;
	i2s_begin(samplerate, 16);

    LPC_GPIO0->FIOSET |= (0x01 << POWER_LED_BIT);	// P0.24 (set high)
}

// I2C->MCP23017->GLCDの初期化
void InitGlcdFull(){
	LPC_GPIO0->FIOCLR |= (0x01 << MCP23017_RESET_PORT);	// P0.6 (MCP23017 reset low)
	I2C1Init();

    LPC_GPIO0->FIOSET |= (0x01 << MCP23017_RESET_PORT);	// P0.6 (set high)
    mcp23017_init();
    glcd_init();
}

// 電源遮断
void PowerOff(){
	SpeakerOn = 0;
	LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_MUTE_BIT);	// P0.26 (set low)
	timer1_counter = 0;
	while (timer1_counter < 4*1){	// 1s

	}
	LPC_GPIO1->FIOCLR |= (0x01 << POWER_ON_BIT);		// P1.31 (set low)
    while(1) {
    }
}

/*
 * This is a private message structure. A generic pointer to this structure
 * is passed to each of the callback functions. Put here any data you need
 * to access from within the callbacks.
 */
struct buffer {
  unsigned char const *start;
  unsigned long length;
};


/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. In this example, an entire file
 * has been mapped into memory, so we just call mad_stream_buffer() with the
 * address and length of the mapping. When this callback is called a second
 * time, we are finished decoding.
 */
static
enum mad_flow input(void *data,
		    struct mad_stream *stream)
{
	static unsigned char strmbuff[INPUT_BUF_SIZE];
	int ret;
	int rsz;
	unsigned char *bp;
	UINT	bytes;
	//int		i;
	//uint16_t	vstep;

	bp = strmbuff;
	rsz = 0;
	if(stream->error == MAD_ERROR_BUFLEN || stream->buffer == 0 /*NULL*/)
	{
	if(stream->next_frame != 0 /*NULL*/)
	  {
		 rsz = stream->bufend - stream->next_frame;
//		 memmove(pStrmBuff, stream->next_frame, rsz);
		 memmove(strmbuff, stream->next_frame, rsz);
		 bp = strmbuff + rsz;
//		 bp = pStrmBuff + rsz;
	  }
	}
	//struct buffer *buffer = data;
	  //if (!buffer->length)

	ret = f_read(pfilMP3, bp, INPUT_BUF_SIZE - rsz, &bytes);
	if (ret != FR_OK || bytes == 0){
		return MAD_FLOW_STOP;
	}

	mad_stream_buffer(stream, strmbuff, bytes + rsz);
//	mad_stream_buffer(stream, pStrmBuff, bytes + rsz);
	//mad_stream_buffer(stream, buffer->start, buffer->length);

	//buffer->length = 0;

	if (AutoPoffTimer >= 180){
		PinEvent = 0;
		return MAD_FLOW_POWEROFF;
	}
	if (PinEvent > 0){
		switch (PinEvent){
		case SWNEXT_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_NEXT;
			break;
		case SWPREV_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_PREV;
			break;
		case SWPLAY_PRESSED:
#if 0
			if (DecodeSkipflg == DECODE_SKIP_OUT){		//曲の途中の無音状態を探してそこまでスキップする。マイコンのスピードが不足して実用的でない。
				DecodeSkipflg = DECODE_SKIP_BEGIN;
				SkipCounter = 0;
			}
#endif
			PinEvent = 0;
			return MAD_FLOW_PLAY;
			break;
		case SWDESELECT_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_DESELECT;
			break;
		case SWOPTION_PRESSED:
			InitGlcdFull();
			DispLcdFullImage();
			/*
			if (glcd_disp_status == 0){
				GLCD_DISP_ON();
			} else {
				GLCD_DISP_OF();
			}
			*/
			break;
		case SWPOFF_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_POWEROFF;
			break;
		}
		PinEvent = 0;
	}
	return MAD_FLOW_CONTINUE;
}

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 * mpg321-0.3.2.origのaudio_linear_ditherを使ってみた。
 */
static inline
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/*
 * mpg321-0.3.2.origから引用
 */
struct audio_dither {
  mad_fixed_t error[3];
  mad_fixed_t random;
};

/*
 * NAME:        prng()
 * DESCRIPTION: 32-bit pseudo-random number generator
 * mpg321-0.3.2.origから引用
 */
static inline
unsigned long prng(unsigned long state)
{
  return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/*
 * NAME:        audio_linear_dither()
 * DESCRIPTION: generic linear sample quantize and dither routine
 * mpg321-0.3.2.origから引用
 */
inline
signed long audio_linear_dither(unsigned int bits, mad_fixed_t sample,
                                struct audio_dither *dither)
{
  unsigned int scalebits;
  mad_fixed_t output, mask, random;

  enum {
    MIN = -MAD_F_ONE,
    MAX =  MAD_F_ONE - 1
  };

  /* noise shape */
  sample += dither->error[0] - dither->error[1] + dither->error[2];

  dither->error[2] = dither->error[1];
  dither->error[1] = dither->error[0] / 2;

  /* bias */
  output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));

  scalebits = MAD_F_FRACBITS + 1 - bits;
  mask = (1L << scalebits) - 1;

  /* dither */
  random  = prng(dither->random);
  output += (random & mask) - (dither->random & mask);

  dither->random = random;

  /* clip */
  if (output > MAX) {
    output = MAX;

    if (sample > MAX)
      sample = MAX;
  }
  else if (output < MIN) {
    output = MIN;

    if (sample < MIN)
      sample = MIN;
  }

  /* quantize */
  output &= ~mask;

  /* error feedback */
  dither->error[0] = sample - output;

  /* scale */
  return output >> scalebits;
}

/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 * ステレオ用
 */
static
enum mad_flow output(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  static struct audio_dither dither;
  dacout_s *dac_w;

  /* pcm->samplerate contains the sampling frequency */
  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  // dac_s2が空くのを待つ(SysTick_Handler参照)
  while(dac_s2 < dac_e2){
	  Pause();
  }

  dac_w = buftab[dacflg];
  dac_s2 = dac_w;
  while (nsamples--) {
		*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*left_ch, VolumeFactor), &dither);
	    dac_w++;
    	*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*right_ch, VolumeFactor), &dither);;
	    dac_w++;
	    left_ch++;
	    right_ch++;
  }
  dac_e2 = dac_w;
  dacflg = 1 - dacflg;

#if 0		// madplayerの原型
//  LPC_GPIO1->FIOSET |= (0x01 << 18);
	enable_timer( 0 );
  while(dac_s < dac_e){
	  Pause();
	  //int k;
	  //for (k=0; k<100; k++);
	  // wait_us(1);		// dacout関数で全データ出力されるのを待つ
  }
	disable_timer( 0 );
//  LPC_GPIO1->FIOCLR |= (0x01 << 18);
  // 音が途切れることは無いのか？
  dac_e = dacbuf;  // potential thread problem ??  no...
  dac_s = dacbuf;

  while (nsamples--) {
	    signed int sample_l,sample_r;
	    sample_l = scale(*left_ch);
	    sample_r = scale(*right_ch);
		*dac_e = (unsigned short)(sample_l);
	    dac_e++;
		*dac_e = (unsigned short)(sample_r);
	    dac_e++;
	    left_ch++;
	    right_ch++;
  }
#endif

  return MAD_FLOW_CONTINUE;
}

/*
 * MONO化output関数
 */
static
enum mad_flow output_monauralize(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
	  unsigned int nsamples;
	  mad_fixed_t const *left_ch, *right_ch;
	  mad_fixed_t	mono_w;
	  static struct audio_dither dither;
	  dacout_s *dac_w;

	  /* pcm->samplerate contains the sampling frequency */
	  //nchannels = pcm->channels;
	  nsamples  = pcm->length;
	  left_ch   = pcm->samples[0];
	  right_ch  = pcm->samples[1];

	  // dac_s2が空くのを待つ(SysTick_Handler参照)
	  while(dac_s2 < dac_e2){
		  Pause();
	  }

	  dac_w = buftab[dacflg];
	  dac_s2 = dac_w;
	  while (nsamples--) {
		 mono_w = *left_ch / 2 + *right_ch / 2;
			*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(mono_w, VolumeFactor), &dither);
			dac_w++;
	    	*dac_w = *(dac_w - 1);
			dac_w++;
			left_ch++;
			right_ch++;
	  }
	  dac_e2 = dac_w;
	  dacflg = 1 - dacflg;

	  return MAD_FLOW_CONTINUE;
}


/*
 * MONO用output関数
 */
static
enum mad_flow output_mono(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  static struct audio_dither dither;
  dacout_s *dac_w;

  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  // dac_s2が空くのを待つ(SysTick_Handler参照)
  while(dac_s2 < dac_e2){
	  Pause();
  }

  dac_w = buftab[dacflg];
  dac_s2 = dac_w;
  while (nsamples--) {
		*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*left_ch, VolumeFactor), &dither);
	    dac_w++;
    	*dac_w = *(dac_w - 1);
	    dac_w++;
	    left_ch++;
	    right_ch++;
  }
  dac_e2 = dac_w;
  dacflg = 1 - dacflg;

  return MAD_FLOW_CONTINUE;
}

#if 0
/*
 * スキップ用
 * 曲の途中の無音状態を探してそこまでスキップする。マイコンのスピードが不足して実用的でない。
 */
static
enum mad_flow output_skip(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  signed int sample_l;
  static struct audio_dither dither;

  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  while (nsamples--) {
	  if (DecodeSkipflg == DECODE_SKIP_IN){
		  sample_l = abs(audio_linear_dither(16, *left_ch, &dither));
		  if (sample_l < 0x0100){			// 無音状態（左チャンネルだけでチェック）
			  SkipCounter++;
			  if (SkipCounter > 44000){	// 無音状態の継続(約1秒）
				  DecodeSkipflg = DECODE_SKIP_END;
				  SkipCounter = 0;
				  break;
			  }
		  } else {
			  SkipCounter = 0;
		  }
	  }
	left_ch++;
	right_ch++;
  }

  return MAD_FLOW_CONTINUE;
}
#endif


/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static
enum mad_flow error(void *data,
		    struct mad_stream *stream,
		    struct mad_frame *frame)
{
#if !defined(MP3_12)
  struct buffer *buffer = data;

  fprintf(stderr, "decoding error 0x%04x (%s) at byte offset %u\n",
	  stream->error, mad_stream_errorstr(stream),
	  stream->this_frame - buffer->start);
#endif

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;
}


/*
 * This is the function called by main() above to perform all the decoding.
 * It instantiates a decoder object and configures it with the input,
 * output, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error).
 * minimad.c(libmad)のdecode関数を元に作成。
 */
static
int decode()
{
//  struct buffer buffer;
  struct mad_decoder decoder;
  int result;

  /* initialize our private message structure */
//  buffer.start  = start;
//  buffer.length = length;

  /* configure input, output, and error functions */
  mad_decoder_init(&decoder, 0 /* &buffer */,
		   input, 0 /* header */, 0 /* filter */, output_monauralize,
		   error, 0 /* message */);

  /* start decoding */
  result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

  /* release the decoder */
  mad_decoder_finish(&decoder);

  return result;
}

/*
 * MP3ファイルごとの初期処理
 * 	output関数
 * 	I2S設定
 * 	SysTick設定
 */
int init_each_mp3(struct mad_decoder *decoder){
	struct mad_header *header;
	header = &(decoder->sync->frame.header);

	// decoder->output_funcの初期値はoutput_monauralize
    if (header->mode == MAD_MODE_SINGLE_CHANNEL){
    	decoder->output_func = output_mono;
    } else if (SpeakerOn == 0 || ChannelMode == 1){		// Speaker Off or Stereo
    	decoder->output_func = output;
    }
    if (samplerate != header->samplerate){	// 設定済みsamplerateと異なる時だけi2s_beginを実行する(POPノイズ対策)
    	if (SpeakerOn == 1){
    		LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_MUTE_BIT);	// Speaker off : P0.26 (set low)
    	}
    	i2s_begin(header->samplerate, 16);
    	samplerate = header->samplerate;
    	if (SpeakerOn == 1){
    		LPC_GPIO0->FIOSET |= (0x01 << SPEAKER_MUTE_BIT);	// speaker on : P0.26 (set high)
    	}
    } else {
    	i2s_start();
    }
	if (SysTick_Config(SystemCoreClock / header->samplerate - 1)) {
	}
	return MAD_FLOW_CONTINUE;
}

#if 0
void skip_begin(struct mad_decoder *decoder){
	output_func_backup = decoder->output_func;
	decoder->output_func = output_skip;
    //LPC_GPIO1->FIOSET |= (0x01 << 20);
}

void skip_end(struct mad_decoder *decoder){
	decoder->output_func = output_func_backup;
    //LPC_GPIO1->FIOCLR |= (0x01 << 20);
}
#endif

/*
 * SysTickタイマー　オフ
 */
void SysTick_Off(){
	SysTick->CTRL &= ~SysTick_CTRL_ENABLE_Msk;
}

/*
 * SysTickタイマー　オン
 */
void SysTick_On(){
	SysTick->CTRL |= SysTick_CTRL_ENABLE_Msk;
}

//  SysTick
void SysTick_Handler(void) {
	if(dac_s < dac_e){
		i2s_write(dac_s, 2);
		dac_s += 2;
	} else if (dac_s2 < dac_e2){
		dac_s = dac_s2;
		dac_e = dac_e2;
		i2s_write(dac_s, 2);
		dac_s += 2;
		dac_s2 = dac_e2;
	}
}

// TIMER0
void TIMER0_IRQHandler (void)
{
	LPC_TIM0->IR = 1;			/* clear interrupt flag */
	timer0_counter++;

	return;
}

void Pause(){
	timer0_counter = 0;
	while (timer0_counter == 0){

	}
}

// TIMER1
void TIMER1_IRQHandler (void){
	//static int	PinEventTmp;
	//static int	ShiftOnTmp;

	LPC_TIM1->IR = 1;			/* clear interrupt flag */
	timer1_counter++;
	if (timer1_counter >= TIMER1_1MIN){
		timer1_counter = 0;
		AutoPoffTimer++;
	}

    if (CHATTERING_TMR == 0){

		if ((LPC_GPIO2->FIOPIN0 & (1 << 5)) == 0){			// P2.5
			if (NowPlaying == 0){
				PinEvent = SWUP_PRESSED;
				if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) == 0) {			// P0.4 (Shift On)
					PinEvent |= SWSHIFT_MASK;
				}
			} else {	// 再生中は音量調整 (この関数内で処理してしまうのでPinEventを設定しない)
		    	if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) != 0) {			// P0.4
					int		i;
					for (i=MIN_VOLUMETAB_IDX; i<=MAX_VOLUMETAB_IDX; i++){	// VolumeTabから直上の値を探す
						if (VolumeTab[i] > Volume){
							Volume = VolumeTab[i];
							break;
						}
					}
				} else {	// Shift+Upで音量微調整
					uint16_t	vstep;
					vstep = Volume / 10;
					if (vstep == 0) vstep = 1;
					if (Volume + vstep <= MAX_VOLUME){
						Volume += vstep;
						//VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
					}
				}
				VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 4)) == 0){	// P2.4
			PinEvent = SWPREV_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 3)) == 0){	// P2.3
			PinEvent = SWNEXT_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 2)) == 0){	// P2.2
			if (NowPlaying == 0){
				PinEvent = SWDOWN_PRESSED;
				if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) == 0) {			// P0.4
					PinEvent |= SWSHIFT_MASK;
				}
			} else {
		    	if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) != 0) {			// P0.4
					int		i;
					for (i=MAX_VOLUMETAB_IDX; i>=MIN_VOLUMETAB_IDX; i--){
						if (VolumeTab[i] < Volume){
							Volume = VolumeTab[i];
							break;
						}
					}
				} else {
					uint16_t	vstep;
					vstep = Volume / 10;
					if (vstep == 0) vstep = 1;
					if (Volume >= vstep){
						Volume -= vstep;
						//VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
					}
				}
				VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO0->FIOPIN1 & (1 << 3)) == 0) {	// P0.11
			PinEvent = SWDESELECT_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 1)) == 0) {	// P2.1
			PinEvent = SWSELECT_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN0 & (1 << 5)) == 0) {	// P0.5
			PinEvent = SWOPTION_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN1 & (1 << 2)) == 0) {	// P0.10
			PinEvent = SWPLAY_PRESSED;
		} else if ((LPC_GPIO1->FIOPIN3 & (1 << 6)) == 0){	// P1.30
			PinEvent = SWPOFF_PRESSED;
		} else {
			PinEvent = 0;
		}

		if (PinEvent != 0){
			CHATTERING_TMR = 1;
		}
    } else if (CHATTERING_TMR > 0){
    	CHATTERING_TMR++;
    	if (CHATTERING_TMR == 3){
    		CHATTERING_TMR = 0;
    	}
    }

}

// TIMER2
void TIMER2_IRQHandler (void){
	LPC_TIM2->IR = 1;			/* clear interrupt flag */
	timer2_counter++;
	onBoardLed1_On();
	//GLCD_DISP_OF();
	glcd_backlight_off();
}

// ミリ秒単位の待ち時間
// TIMER0を使うので再生中は使えない。
void WaitMs(int ms){
	ms *= 10;
	timer0_counter = 0;
	while (timer0_counter < ms){

	}
}

// 選択画面における入力待ち
int PauseAndSelectItem(){
	int		ret;

	//DebugWriteln("PauseAndSelectItem");
	ret = 0;
	PinEvent = 0;
	AutoPoffTimer = 0;
	do {
		if (PinEvent != 0){
			ret = PinEvent;
			//DebugWritelnV1(" ret=", ret);
			PinEvent = 0;
			return ret;
		}
		if (AutoPoffTimer == 2){
			return SWPOFF_PRESSED;
		}
	} while (1);
}


// メインメニューまたはオプションメニューのイメージデータを取得
int ReadMainOrOptionMenuImage(FIL *Fil, uint8_t menukind){
    FRESULT     res;
    UINT		br;

    res = f_open(Fil, MENUDATFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        return FALSE;
    }

    if (menukind == COM_KND_MAINMENU){
    } else if (menukind == COM_KND_OPTIONMENU) {
		res = f_lseek(Fil, LCD_IMAGE_BUFFER_SIZE);		// MAINMENUのデータを読み飛ばす
    } else {
    	return FALSE;
    }

	memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
	res = f_read(Fil, LcdImageBuff, LCD_IMAGE_BUFFER_SIZE, &br);

    f_close(Fil);
	glcd_backlight_on();

    return TRUE;
}

/*
 * ディレクトリの選択操作
 * 	レコード番号：ヘッダレコードを０とする。以降、ルートディレクトリを１、など。
 * 	再生中のボタン機能
 * 	NEXTボタン：同一フォルダの次のトラックへ
 * 	PLAYボタン：ディレクトリ選択へ戻る。
 * 	DESELECTボタン：同一ディレクトリのトラック選択に戻る
 */
int SelectDirectory(uint8_t *RecDirList, FIL *Fil){
    FRESULT     res;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;			// 現在対象のディレクトリの先頭サブディレクトリのレコード番号
    uint16_t    numSubDir;
    uint16_t    idxSubDir;				// 選択ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurSubDirItem;   	// 現在表示ページで選択中のDIR(0,1,2,3,...,9)
    uint8_t     idxCurSubDirItemBk;		// 上記の一時的退避
    uint8_t		NumSelItem;				// １ページに表示する最大項目数
    //uint8_t		MaxSelItem;				// 上記最大数
    int        	ret;
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[MAX_SELDIR_ITEM];
    uint16_t    numSubDirTab[MAX_SELDIR_ITEM];
    uint16_t    numSndFile[MAX_SELDIR_ITEM];
    uint16_t    recNoFirstSubdir[MAX_SELDIR_ITEM];
    uint8_t     readdlflg;      // DIR LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;

	//DebugWritelnV1("SelectDirectory start ", u8ControllerVer);
    ret = FR_OK;

    res = f_open(Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
    	//DebugWriteln("f_open error\r\n");
        return res;
    }

    // Skip HEADER info
    res = f_lseek(Fil, RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	//DebugWriteln("f_lseek error\r\n");
        return res;
    }

    // Read ROOT info
    res = f_read_DIRLISTFILE(Fil, RecDirList);
    if (res != FR_OK){
    	//DebugWriteln("f_read_DIRLISTFILE error\r\n");
        return res;
    }

	memcpy(&rowSubDirStart, RecDirList + DL_RECNO_SUBDIR_START, 2);				// OK
	//DebugWritelnV1("f_read_DIRLISTFILE 2 rowSubDirStart ", rowSubDirStart);
    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
    numSubDir = *(RecDirList + DL_NUM_SUBDIR);
    idxSubDir = 0;
    idxCurSubDirItem = 0;
    idxCurSubDirItemBk = -1;
    idxDirDepth = 0;
    readdlflg = TRUE;
    NumSelItem = MAX_SELDIR_ITEM;

    // end point of ROOT info
    res = f_lseek(Fil, 2*RECLEN_DIRLIST_FULL);		// ヘッダレコードとルートディレクトリレコードを読み飛ばす
    if (res != FR_OK){
    	//DebugWriteln("f_lseek error\r\n");
        return res;
    }

    while (1){
    	if (readdlflg == TRUE){
    		items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			res = f_lseek(Fil, ofsSubDirStart + (uint32_t)idxSubDir * RECLEN_DIRLIST_FULL);
			for (i=0; i<NumSelItem; i++){
				if ((idxSubDir + i) < numSubDir){
					if ((res = f_read_DIRLISTFILE(Fil, RecDirList)) == FR_OK){

						dirno[i] = rowSubDirStart + idxSubDir + i - 1;
                        numSubDirTab[i] = *(RecDirList + DL_NUM_SUBDIR);
						numSndFile[i] = *(RecDirList + DL_NUM_SNDFILE);
						memcpy(&recNoFirstSubdir[i], RecDirList + DL_RECNO_SUBDIR_START, 2);
						items++;
						SetRecDirList2MenuImage(RecDirList + RECLEN_DIRLIST, i, NumSelItem);
					} else {
						// エラー
					}
				} else {
					dirno[i] = 0;
					numSndFile[i] = 0;
					recNoFirstSubdir[i] = 0;
				}
			}
			SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
        	DispLcdFullImage();
    	}

		//DebugWritelnV1("idxCurrentSubDirItem=", idxCurrentSubDirItem);
		ret = PauseAndSelectItem();
//		LedOneBlink();
	    readdlflg = TRUE;
		//DebugWritelnV1("PauseAndSelectItem=", ret);
		if (ret == SWNEXT_PRESSED){
			// 同一行内の項目を順次選択 ： 同一行の右の項目を選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem + LCD_STRING_LINES) < items){
				idxCurSubDirItem += LCD_STRING_LINES;
			} else {
				idxCurSubDirItem %= LCD_STRING_LINES;
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsLeftRight(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
			readdlflg = FALSE;
		} else if (ret == SWPREV_PRESSED){
			// 同一行内の項目を順次選択 ： 同一行の右の項目を選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxCurSubDirItem >= LCD_STRING_LINES){
				idxCurSubDirItem -= LCD_STRING_LINES;
			} else {
				while (1){
					if ((idxCurSubDirItem + LCD_STRING_LINES) < items){
						idxCurSubDirItem += LCD_STRING_LINES;
					} else {
						break;
					}
				}
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsLeftRight(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
			readdlflg = FALSE;
		} else if (ret == SWUP_PRESSED){
			// 同一ページ内の上の行へ、または、前ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == 0){
				if (idxSubDir >= NumSelItem){
					idxSubDir -= NumSelItem;	// 前ページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = FALSE;
				}
			} else {
				idxCurSubDirItem--;
				readdlflg = FALSE;
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsUpDown(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){
			// 同一ページ内の下の行へ、または、次ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == (LCD_STRING_LINES - 1)){
				if ((idxSubDir + NumSelItem) < numSubDir){
					idxSubDir += NumSelItem;	// 次のページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = FALSE;
				}
			} else {
				if (idxCurSubDirItem < (items - 1)){
					idxCurSubDirItem++;
					readdlflg = FALSE;
				} else {
					readdlflg = FALSE;
				}
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsUpDown(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
		} else if (ret == SWUPSHIFT_PRESSED){		// 改ページ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxSubDir >= NumSelItem){
				idxSubDir -= NumSelItem;	// 前ページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWDOWNSHIFT_PRESSED){	// 改ページ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxSubDir + NumSelItem) < numSubDir){
				idxSubDir += NumSelItem;	// 次のページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPLAY_PRESSED){		// PLAY
			if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				res = f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL);
				res = f_read_DIRLISTFILE(Fil, RecDirList);

				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_PLAY);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
			if (recNoFirstSubdir[idxCurSubDirItem] > 1){		// 下位DIRが存在する
				rowSubDirStart = recNoFirstSubdir[idxCurSubDirItem];
				ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
				numSubDir = numSubDirTab[idxCurSubDirItem];		// *(RecDirList + DL_NUM_SUBDIR);
				// push current idxSubDir to stack
				stackIdxSubDir[idxDirDepth] = idxSubDir;
				stackIdxCurrentSubDirItem[idxDirDepth] = idxCurSubDirItem;
				idxDirDepth++;
				idxSubDir = 0;
				idxCurSubDirItem = 0;
				//DebugWritelnV1("rowSubDirStart=", rowSubDirStart);
		        //NumSelItem = MAX_SELDIR_ITEM;		//初期値に戻さない
			} else if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				res = f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL);
				res = f_read_DIRLISTFILE(Fil, RecDirList);

				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_SELECT);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWDESELECT_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
			uint16_t    r;
			uint16_t    r2;

			//r = *((uint16_t *)(RecDirList + DL_RECNO_GRANDPARENTDIR));   // rec no of grand parent
			//r = P8U16(RecDirList, DL_RECNO_GRANDPARENTDIR);		// rec no of grand parent
			memcpy(&r,  RecDirList + DL_RECNO_GRANDPARENTDIR, 2);	// rec no of grand parent
			memcpy(&r2, RecDirList + DL_RECNO_PARENTDIR, 2);
			// 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
			if (r == r2){
				ret = GOTO_MAINMENU;
				break;
			} else {
				// Go to grand parent record
				res = f_lseek(Fil, (uint32_t)r * RECLEN_DIRLIST_FULL);
				if (res != FR_OK) return res;
				// Read grand parent record
				res = f_read_DIRLISTFILE(Fil, RecDirList);
				if (res != FR_OK) return res;
				// the first "uncle叔父" record
				//r = *((uint16_t *)(RecDirList + DL_RECNO_SUBDIR_START));
				//r = P8U16(RecDirList, DL_RECNO_SUBDIR_START);
				memcpy(&r, RecDirList + DL_RECNO_SUBDIR_START, 2);
				if (r > 1){
					rowSubDirStart = r;
					ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
					numSubDir = *(RecDirList + DL_NUM_SUBDIR);
					// pop idxSubDir from stack -> This is parent
					idxDirDepth--;
					idxSubDir = stackIdxSubDir[idxDirDepth];
					idxCurSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
					if (idxSubDir >= numSubDir) idxSubDir = 0;
			        //NumSelItem = MAX_SELDIR_ITEM;		// 初期値に戻さない
				} else {
					ret = GOTO_MAINMENU;
					break;
				}
			}
			readdlflg = TRUE;
		} else if (ret == SWOPTION_PRESSED){      //
		    f_close(Fil);
			ret = SelectOptionMenu(Fil);
		    f_open(Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
			readdlflg = TRUE;
			if (ret == OPTION_1){
				InitGlcdFull();
			} else if (ret == OPTION_2){
				ToggleSpeakerOnOff();
			} else if (ret == OPTION_3){
				ToggleChannelModeOnOff();
			} else if (ret == POWEROFF){
				ret = POWEROFF;
				break;
			}
		} else if (ret == SWNEXTSHIFT_PRESSED){		// 一度に表示できるメニュー項目数を減らす
			if (NumSelItem == MAX_SELDIR_ITEM){
				NumSelItem = MAX_SELDIR_ITEM/2;
				if (idxCurSubDirItem >= LCD_STRING_LINES){
					idxCurSubDirItem -= LCD_STRING_LINES;
				}
			} else {
			    readdlflg = FALSE;
			}
		} else if (ret == SWPREVSHIFT_PRESSED){		// 一度に表示できるメニュー項目数を増やす
			if (NumSelItem == MAX_SELDIR_ITEM/2){
				NumSelItem = MAX_SELDIR_ITEM;
			} else {
			    readdlflg = FALSE;
			}
		} else if (ret == SWPOFF_PRESSED){
			ret = POWEROFF;
			break;
		}

    }

    f_close(Fil);
    return ret;
}

/*
 * ディレクトリリストファイルの基本部分+LCD表示イメージを読み込む
 */
FRESULT f_read_DIRLISTFILE (FIL* fp, void* buff){
    FRESULT res1;
    UINT    br;
    res1 = f_read(fp, buff, RECLEN_DIRLIST_FULL, &br);
    if (br != RECLEN_DIRLIST_FULL) return FR_NO_FILE;
    return res1;
}

/*
 * インデックスの選択操作
 */
int SelectIndex(uint8_t *RecDirList, FIL *Fil){
    FRESULT     res;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;
    uint16_t    numSubDir;
    uint16_t    idxSubDir;				// 選択ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurSubDirItem;   	// 現在表示ページで選択中のDIR(0,1,2,3)
    uint8_t     idxCurSubDirItemBk;		// 上記の一時的退避
    uint8_t		NumSelItem;				// １ページに表示する項目数
    //uint8_t		MaxSelItem;				// 上記最大数
    int        	ret;
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[NUM_SELINDEX_ITEM];
    uint16_t    numSubDirTab[NUM_SELINDEX_ITEM];
    uint16_t    numSndFile[NUM_SELINDEX_ITEM];
    uint16_t    recNoFirstSubdir[NUM_SELINDEX_ITEM];
    uint8_t     readdlflg;      // DIR LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;
    uint8_t		LcdStringLines;

	//DebugWriteln("SelectDirectory start\r\n");
    ret = FR_OK;

    res = f_open(Fil, INDEXFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
    	//DebugWriteln("f_open error\r\n");
        return res;
    }

    // Skip HEADER info
    res = f_lseek(Fil, RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
	//DebugWriteln("f_lseek error\r\n");
        return res;
    }
	//DebugWriteln("f_lseek ok\r\n");

    // Read ROOT info
    res = f_read_DIRLISTFILE(Fil, RecDirList);
    if (res != FR_OK){
    	//DebugWriteln("f_read_DIRLISTFILE error\r\n");
        return res;
    }
	//DebugWriteln("f_read_DIRLISTFILE 1\r\n");

	memcpy(&rowSubDirStart, RecDirList + DL_RECNO_SUBDIR_START, 2);				// ルートレコードからサブディレクトリの先頭レコード番号を取得
	//DebugWritelnV1("f_read_DIRLISTFILE 2 rowSubDirStart ", rowSubDirStart);
    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
    numSubDir = *(RecDirList + DL_NUM_SUBDIR);							// ルートディレクトリからサブディレクトリ数（インデックス数）を取得
    idxSubDir = 0;
    idxCurSubDirItem = 0;
    idxCurSubDirItemBk = -1;
    idxDirDepth = 0;
    readdlflg = TRUE;
    NumSelItem = NUM_SELINDEX_ITEM;		// インデックス選択の項目数（4行 X １行あたり１０項目）
    LcdStringLines = SELINDEX_LINES;	//　インデックス選択の行数　GetLcdStringLines();

    // end point of ROOT info
    res = f_lseek(Fil, 2*RECLEN_DIRLIST_FULL);		// ヘッダレコードとルートディレクトリレコードを読み飛ばす
    if (res != FR_OK){
	//DebugWriteln("f_lseek error\r\n");
        return res;
    }
	//DebugWriteln("f_lseek ok\r\n");

    while (1){
    	if (readdlflg == TRUE){
    		items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			res = f_lseek(Fil, ofsSubDirStart + (uint32_t)idxSubDir * RECLEN_DIRLIST_FULL);
			for (i=0; i<NumSelItem; i++){
				if ((idxSubDir + i) < numSubDir){
					if ((res = f_read_DIRLISTFILE(Fil, RecDirList)) == FR_OK){

						dirno[i] = rowSubDirStart + idxSubDir - 1;
                        numSubDirTab[i] = *(RecDirList + DL_NUM_SUBDIR);
						numSndFile[i] = *(RecDirList + DL_NUM_SNDFILE);
						memcpy(&recNoFirstSubdir[i], RecDirList + DL_RECNO_SUBDIR_START, 2);
						items++;
						if (idxDirDepth == 0){	// インデックスのトップページ(A,B,C,...フォルダの頭文字順)
							SetRecDirList2IndexMenuImage(RecDirList + RECLEN_DIRLIST, i);
						} else {	// インデックスのサブページ（各アルバム・作品などMP3ファイルを持つフォルダ)
							SetRecDirList2MenuImage(RecDirList + RECLEN_DIRLIST, i, NumSelItem);
						}
					} else {
						// エラー
					}
				} else {
					dirno[i] = 0;
					numSndFile[i] = 0;
					recNoFirstSubdir[i] = 0;
				}
			}
			SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
        	DispLcdFullImage();
    	}

		//DebugWritelnV1("idxCurrentSubDirItem=", idxCurSubDirItem);
		ret = PauseAndSelectItem();
		//LedOneBlink();
	    readdlflg = TRUE;
		//DebugWritelnV1("PauseAndSelectItem=", ret);
		if (ret == SWNEXT_PRESSED){
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem + LcdStringLines /*LCD_STRING_LINES*/) < items){
				idxCurSubDirItem += LcdStringLines /*LCD_STRING_LINES*/;
			} else {
				idxCurSubDirItem %= LcdStringLines /*LCD_STRING_LINES*/;
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsLeftRight(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
			readdlflg = FALSE;
		} else if (ret == SWPREV_PRESSED){      // 同一ページの前項目選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxCurSubDirItem >= LcdStringLines /*LCD_STRING_LINES*/){
				idxCurSubDirItem -= LcdStringLines /*LCD_STRING_LINES*/;
			} else {
				while (1){
					if ((idxCurSubDirItem + LcdStringLines /*LCD_STRING_LINES*/) < items){
						idxCurSubDirItem += LcdStringLines /*LCD_STRING_LINES*/;
					} else {
						break;
					}
				}
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsLeftRight(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
			readdlflg = FALSE;
		} else if (ret == SWUP_PRESSED){        // 同一DIRの前ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LcdStringLines /*LCD_STRING_LINES*/) == 0){
				if (idxSubDir >= NumSelItem){
					idxSubDir -= NumSelItem;
					idxCurSubDirItem = 0;
				} else {
					readdlflg = FALSE;
				}
			} else {
				idxCurSubDirItem--;
				readdlflg = FALSE;
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsUpDown(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){      // 同一DIRの次ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LcdStringLines /*LCD_STRING_LINES*/) == (LcdStringLines /*LCD_STRING_LINES*/ - 1)){
				if ((idxSubDir + NumSelItem) < numSubDir){
					idxSubDir += NumSelItem;
					idxCurSubDirItem = 0;
				} else {
					readdlflg = FALSE;
				}
			} else {
				if (idxCurSubDirItem < (items - 1)){
					idxCurSubDirItem++;
					readdlflg = FALSE;
				} else {
					readdlflg = FALSE;
				}
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
				DispMenuItemsUpDown(idxCurSubDirItem, idxCurSubDirItemBk, NumSelItem);
	        	//DispLcdFullImage();
			}
		} else if (ret == SWPLAY_PRESSED){             // PLAY
			if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				res = f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL);
				res = f_read_DIRLISTFILE(Fil, RecDirList);

				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_PLAY);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
			if (recNoFirstSubdir[idxCurSubDirItem] > 1){		// 下位DIRが存在する
				rowSubDirStart = recNoFirstSubdir[idxCurSubDirItem];
				ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
				numSubDir = numSubDirTab[idxCurSubDirItem];		// *(RecDirList + DL_NUM_SUBDIR);
				// push current idxSubDir to stack
				stackIdxSubDir[idxDirDepth] = idxSubDir;
				stackIdxCurrentSubDirItem[idxDirDepth] = idxCurSubDirItem;
				idxDirDepth++;
				idxSubDir = 0;
				idxCurSubDirItem = 0;
				//DebugWritelnV1("rowSubDirStart=", rowSubDirStart);
			    LcdStringLines = LCD_STRING_LINES;
		        NumSelItem = MAX_SELDIR_ITEM;
			} else if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				res = f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL);
				res = f_read_DIRLISTFILE(Fil, RecDirList);

				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_SELECT);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWDESELECT_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
			uint16_t    r;
			uint16_t    r2;

			//r = *((uint16_t *)(RecDirList + DL_RECNO_GRANDPARENTDIR));   // rec no of grand parent
			//r = P8U16(RecDirList, DL_RECNO_GRANDPARENTDIR);		// rec no of grand parent
			memcpy(&r,  RecDirList + DL_RECNO_GRANDPARENTDIR, 2);	// rec no of grand parent
			memcpy(&r2, RecDirList + DL_RECNO_PARENTDIR, 2);
			// 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
			if (r == r2){
				ret = GOTO_MAINMENU;
				break;
			} else {
				// Go to grand parent record
				res = f_lseek(Fil, (uint32_t)r * RECLEN_DIRLIST_FULL);
				if (res != FR_OK) return res;
				// Read grand parent record
				res = f_read_DIRLISTFILE(Fil, RecDirList);
				if (res != FR_OK) return res;
				// the first "uncle叔父" record
				//r = *((uint16_t *)(RecDirList + DL_RECNO_SUBDIR_START));
				//r = P8U16(RecDirList, DL_RECNO_SUBDIR_START);
				memcpy(&r, RecDirList + DL_RECNO_SUBDIR_START, 2);
				if (r > 1){
					rowSubDirStart = r;
					ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
					numSubDir = *(RecDirList + DL_NUM_SUBDIR);
					// pop idxSubDir from stack -> This is parent
					idxDirDepth--;
					idxSubDir = stackIdxSubDir[idxDirDepth];
					idxCurSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
					if (idxSubDir >= numSubDir) idxSubDir = 0;
				    LcdStringLines = SELINDEX_LINES;
			        NumSelItem = NUM_SELINDEX_ITEM;
				} else {
					ret = GOTO_MAINMENU;
					break;
				}
			}
			readdlflg = TRUE;
		} else if (ret == SWOPTION_PRESSED){      //
		    f_close(Fil);
			ret = SelectOptionMenu(Fil);
		    f_open(Fil, INDEXFILE, FA_READ | FA_OPEN_EXISTING);
			readdlflg = TRUE;
			if (ret == OPTION_1){
				InitGlcdFull();
			} else if (ret == OPTION_2){
				ToggleSpeakerOnOff();
			} else if (ret == OPTION_3){
				ToggleChannelModeOnOff();
			} else if (ret == POWEROFF){
				ret = POWEROFF;
				break;
			}
		} else if (ret == SWPOFF_PRESSED){
			ret = POWEROFF;
			break;
		}

    }

    f_close(Fil);
    return ret;
}

// 現在のサブディレクトリからMP3を選択する。または再生開始する。
int PlayOrSelectMp3InSubdirectory(uint8_t *RecDirList, uint16_t dirno, uint16_t numsndfile, uint8_t start_trackno, playorselect_t playorselect){
    FRESULT     res;
    int		    ret;
    FIL         filML;		// MP3リストファイル用
    uint8_t     RecMp3List[RECLEN_MP3LIST_FULL];   // for MP3 list file and VOLUME file
    uint8_t     trackno;		// 再生中の曲番 (0,1,2,...,numsndfile-1)
    uint8_t     idxtrack;       // 選択操作中のページ先頭の曲番 (0,1,2,...,numsndfile-1)
    uint8_t		NumSelItem;		// １ページに表示する項目数
    UINT        br;
    uint16_t 	volinit = 0;
    uint8_t     dirlen;
    uint8_t     execplayflg = 0;
    uint8_t     readdlflg;      // MP3 LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;
    uint16_t    idxCurTitleItem;
    uint16_t    idxCurTitleItemBk;

    ret = FR_OK;
    dirlen = *(RecDirList + DL_LENFULLDIR);

//	DebugWriteln((char *)(RecDirList + DL_FULLDIR));

    // Volume file
    *(RecDirList + DL_FULLDIR + dirlen) = '/';
    strcpy(RecDirList + DL_FULLDIR + dirlen + 1, VOLUMEFILE);
    res = f_open(&filML, (TCHAR *)(RecDirList + DL_FULLDIR), FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        if (f_read(&filML, RecMp3List, RECLEN_VOLUME, &br) == FR_OK && br == RECLEN_VOLUME){
			//volinit = RecMp3List[VF_VOLUME];
        	if (SpeakerOn == 1){
        		memcpy(&volinit, &RecMp3List[VF_VOLUME], sizeof(volinit));
        	} else {
        		memcpy(&volinit, &RecMp3List[VF_VOLUME_2], sizeof(volinit));
        	}
			//DebugWritelnV1("get volfile ", volinit);
			if (volinit <= MAX_VOLUME && volinit > MIN_VOLUME){
				Volume = volinit;
			}
        }
        f_close(&filML);
    }
    if (!(Volume <= MAX_VOLUME && Volume > MIN_VOLUME)){
    	Volume = VolumeTab[DEF_VOLUMETAB_IDX];
    }
    //Volume = VolumeTab[0];		// for TEST
	VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);

    // MP3 file list
    *(RecDirList + DL_FULLDIR + dirlen) = '/';
    strcpy(RecDirList + DL_FULLDIR + dirlen + 1, SNDFILELISTFILE);
    res = f_open(&filML, (TCHAR *)(RecDirList + DL_FULLDIR), FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        return res;
    }

    if (start_trackno >= numsndfile) start_trackno = 0;
    trackno = start_trackno;
    idxtrack = 0;
    idxCurTitleItem = 0;
    readdlflg = TRUE;
    NumSelItem = MAX_SELMP3_ITEM;

    while (1){
        if (PlayMode == PM_SELECT && playorselect == PS_SELECT){	// 選択操作を行う
        	if (readdlflg == TRUE){		//
        		// 選択肢を表示
        		items = 0;
    			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
				res = f_lseek(&filML, (uint32_t)idxtrack * RECLEN_MP3LIST_FULL);
        		for (i=0; i<NumSelItem; i++){
        			if ((idxtrack + i) < numsndfile){
        				if ((res = f_read(&filML, RecMp3List, RECLEN_MP3LIST_FULL, &br)) == FR_OK){
    						items++;
  							SetRecDirList2MenuImage(RecMp3List + RECLEN_MP3LIST, i, NumSelItem);
    					} else {
    						// エラー
    					}
        			} else {
    					//memset(&(OledImageWork[i*SELMP3_ITEM_WIDTH]), 0, SELMP3_ITEM_WIDTH);
        			}
        		}
    			SetMenuItemSelected(idxCurTitleItem, NumSelItem);
            	DispLcdFullImage();
        	}

            ret = PauseAndSelectItem();
            //LedOneBlink();
    	    readdlflg = TRUE;
            if (ret == SWNEXT_PRESSED){          // NEXT mp3 file
    			readdlflg = FALSE;
                continue;		// 選択操作続行
            } else if (ret == SWPREV_PRESSED){	// PREV mp3 file
    			readdlflg = FALSE;
                continue;
            } else if (ret == SWUP_PRESSED){
            	idxCurTitleItemBk = idxCurTitleItem;
            	if (idxCurTitleItem == 0){
        			//idxCurTitleItem = NumSelItem - 1;
                	if (idxtrack >= NumSelItem){
                		idxtrack -= NumSelItem;
                		//idxCurTitleItem = 0;
                		idxCurTitleItem = NumSelItem - 1;
                	} else {
            			readdlflg = FALSE;
                	}
            	} else {
            		idxCurTitleItem--;
        			readdlflg = FALSE;
            	}
    			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
    				SetMenuItemSelected(idxCurTitleItemBk, NumSelItem);
    				SetMenuItemSelected(idxCurTitleItem, NumSelItem);
    	        	DispLcdFullImage();
    			}
                continue;
            } else if (ret == SWDOWN_PRESSED){
            	idxCurTitleItemBk = idxCurTitleItem;
				if (idxCurTitleItem < (NumSelItem - 1)){
					if ((idxtrack + idxCurTitleItem) < (numsndfile - 1)){
						idxCurTitleItem++;
						readdlflg = FALSE;
					} else {
	        			readdlflg = FALSE;
					}
				} else {
					//idxCurTitleItem = 0;
	            	if ((idxtrack + NumSelItem) < numsndfile){
	            		idxtrack += NumSelItem;
	            		idxCurTitleItem = 0;
	            	} else {
	        			readdlflg = FALSE;
	            	}
				}
    			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
    				SetMenuItemSelected(idxCurTitleItemBk, NumSelItem);
    				SetMenuItemSelected(idxCurTitleItem, NumSelItem);
    	        	DispLcdFullImage();
    			}
            	continue;
    		} else if (ret == SWUPSHIFT_PRESSED){		// 改ページ
    			if (idxtrack >= NumSelItem){
    				idxtrack -= NumSelItem;
    				idxCurTitleItem = 0;
    			} else {
    				readdlflg = FALSE;
    			}
				continue;
    		} else if (ret == SWDOWNSHIFT_PRESSED){	// 改ページ
    			if ((idxtrack + NumSelItem) < numsndfile){
    				idxtrack += NumSelItem;
    				idxCurTitleItem = 0;
    			} else {
    				readdlflg = FALSE;
    			}
				continue;
            } else if (ret == SWPLAY_PRESSED){
            	//DebugWriteln("SWPLAY_PRESSED pressed \r\n");
                playorselect = PS_PLAY;		// 再生開始
                trackno = idxtrack + idxCurTitleItem;
        		NumSelItem = MAX_SELMP3_ITEM;
                continue;		// いったんループの先頭に戻る
            } else if (ret == SWDESELECT_PRESSED){
            	ret = PLAY_OTHER_DIR;
        		NumSelItem = MAX_SELMP3_ITEM;
                break;			// 現ディレクトリを抜ける。
            } else if (ret == SWOPTION_PRESSED){
    		    f_close(&filML);
    			ret = SelectOptionMenu(&filML);
    		    *(RecDirList + DL_FULLDIR + dirlen) = '/';
    		    strcpy(RecDirList + DL_FULLDIR + dirlen + 1, SNDFILELISTFILE);
    		    res = f_open(&filML, (TCHAR *)(RecDirList + DL_FULLDIR), FA_READ | FA_OPEN_EXISTING);
    			readdlflg = TRUE;
    			if (ret == OPTION_1){
    				InitGlcdFull();
    			} else if (ret == OPTION_2){
    				ToggleSpeakerOnOff();
    			} else if (ret == OPTION_3){
    				ToggleChannelModeOnOff();
    			} else if (ret == POWEROFF){
    				ret = POWEROFF;
    				break;
    			}
            	continue;
            } else if (ret == SWPOFF_PRESSED){
				ret = POWEROFF;
				break;
            } else {
    			readdlflg = FALSE;
                continue;
            }
        } else {	// 再生開始
        	//onBoardLed1_Off();
        	SetImageAtPlaying(&filML, RecMp3List, trackno, numsndfile, NumSelItem);
        	DispLcdFullImage();
        	glcd_backlight_on();
        	//enable_timer( 2 );		// 所定時間後バックライトを消す ->音が途切れるので廃案
            if (f_lseek(&filML, (uint32_t)(trackno)*(RECLEN_MP3LIST_FULL)) != FR_OK) break;
            if (f_read(&filML, RecMp3List, RECLEN_MP3LIST_FULL, &br) != FR_OK) break;
        }

        // Mp3 file full path
        *(RecDirList + DL_FULLDIR + dirlen) = '/';
        strcpy(RecDirList + DL_FULLDIR + dirlen + 1, RecMp3List + ML_FILENAME);
        // Play a MP3 file
        ret = PlayOneMp3((char *)(RecDirList + DL_FULLDIR), dirno, trackno /*, Fil*/);
        //DebugWritelnV1("PlayOneMp3=", ret);
        if (ret == PLAY_OTHER_DIR){
            break;
        } else if (ret == POWEROFF){
            break;
        } else if (ret == GOTO_MAINMENU){
            break;
        } else if (ret == SELECT_OTHER_MP3){
            if (PlayMode == PM_RANDOM_DIR){
            	ret = GOTO_MAINMENU;
            	break;
            } else {
                playorselect = PS_SELECT;
            }
        } else if (ret == PLAY_PREV_MP3){
            if (trackno > 0) trackno--;
        } else {
            execplayflg = 1;
            trackno++;
            if (trackno == numsndfile) break;
        }

    }
    f_close(&filML);

    // VOLUMEFILE保存
    if (execplayflg == 1 && Volume != volinit){     // at least one MP3 was played
        *(RecDirList + DL_FULLDIR + dirlen) = '/';
        strcpy(RecDirList + DL_FULLDIR + dirlen + 1, VOLUMEFILE);
        res = f_open(&filML, (TCHAR *)(RecDirList + DL_FULLDIR), FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
        if (res == FR_OK){
            if (f_read(&filML, RecMp3List, RECLEN_VOLUME, &br) != FR_OK || br != RECLEN_VOLUME){
            	RecMp3List[VF_COUNT] = 0;
            }
        	if (SpeakerOn == 1){
        		memcpy(&RecMp3List[VF_VOLUME], &Volume, sizeof(Volume));
        	} else {
        		memcpy(&RecMp3List[VF_VOLUME_2], &Volume, sizeof(Volume));
        	}
			RecMp3List[VF_COUNT]++;
			f_lseek(&filML, 0);
			f_write(&filML, RecMp3List, RECLEN_VOLUME, &br);
            f_close(&filML);
        }
    }

    return ret;
}

/*
 * オプションメニュー選択操作
 */
int SelectOptionMenu(FIL *Fil){
    int         return_to_pauseandselect;
    uint8_t		selectedmenuitem = 0;
    uint8_t		selectedmenuitemBk;
	int			ret;
    uint8_t     readdlflg;

	return_to_pauseandselect = TRUE;
	readdlflg = TRUE;

	do {
		if (readdlflg == TRUE){
			ReadMainOrOptionMenuImage(Fil, COM_KND_OPTIONMENU);
			SetMenuItemSelected(selectedmenuitem, LCD_ITEMS_MODE_5);
			DispLcdFullImage();
			readdlflg = FALSE;
		}

		// オプションメニュー入力待ち
		ret = PauseAndSelectItem();
		switch (ret){
		case SWDOWN_PRESSED:
			selectedmenuitemBk = selectedmenuitem;
			if (selectedmenuitem < (NUM_OPTIONMENU - 1)){
				selectedmenuitem++;
			} else {
				selectedmenuitem = 0;
			}
			if (selectedmenuitem != selectedmenuitemBk){
				SetMenuItemSelected(selectedmenuitemBk, LCD_ITEMS_MODE_5);
				SetMenuItemSelected(selectedmenuitem, LCD_ITEMS_MODE_5);
				DispMenuItemsUpDown(selectedmenuitem, selectedmenuitemBk, LCD_ITEMS_MODE_5);
				readdlflg = FALSE;
			}
			break;
		case SWUP_PRESSED:
			selectedmenuitemBk = selectedmenuitem;
			if (selectedmenuitem > 0){
				selectedmenuitem--;
			} else {
				selectedmenuitem = NUM_OPTIONMENU - 1;
			}
			if (selectedmenuitem != selectedmenuitemBk){
				SetMenuItemSelected(selectedmenuitemBk, LCD_ITEMS_MODE_5);
				SetMenuItemSelected(selectedmenuitem, LCD_ITEMS_MODE_5);
				DispMenuItemsUpDown(selectedmenuitem, selectedmenuitemBk, LCD_ITEMS_MODE_5);
				readdlflg = FALSE;
			}
			break;
		case SWSELECT_PRESSED:
		case SWOPTION_PRESSED:
			switch (selectedmenuitem){
			case OPTIONMENU_1:
				ret = OPTION_1;
				break;
			case OPTIONMENU_2:
				ret = OPTION_2;
				break;
			case OPTIONMENU_3:
				ret = OPTION_3;
				break;
			default:
				ret = DO_NOTHING;
				break;
			}
			return_to_pauseandselect = FALSE;
			break;
		case SWDESELECT_PRESSED:
			ret = DO_NOTHING;
			return_to_pauseandselect = FALSE;
			break;
		case SWPOFF_PRESSED:
			ret = POWEROFF;
			return_to_pauseandselect = FALSE;
			break;
		default:
			break;
		}
	} while (return_to_pauseandselect);

	return ret;
}

/*
 * ディレクトリリストファイル/MP3リストファイル のレコードからディレクトリ選択メニューイメージを作成する。
 * 	no : メニュー項目の序数 (0 - 9)
 * 	itemsperpage : 一度に表示する項目数(10 or 8 or 5 or 4)
 */
void SetRecDirList2MenuImage(uint8_t *RecDirList, uint8_t no, uint8_t itemsperpage){
	static uint8_t	no2page_10[10] = {0, 1, 3, 4, 6, 0, 1, 3, 4, 6};	// itemsperpage=10/5のとき、各noのイメージをセットするページ(no=0は0と1、no=1は1と2、no=2は3と4)
	static uint8_t	halfflg_10[10] = {0, 1, 0, 1, 0, 0, 1, 0, 1, 0};	// 0はnoの項目が１ページすべてと次のページの上半に渡ってセットされる。1はnoの項目がページの下半と次のページすべてに渡ってセットされる
	static uint8_t	no2page_8[8]   = {0, 1, 3, 4, 0, 1, 3, 4};			// itemsperpage=8/4のとき、同上（各noは２つのページを使う。）
	static uint8_t	halfflg_8[8]   = {0, 1, 0, 1, 0, 1, 0, 1};			//

	int		x, page;	// pageはLCDのドキュメントで言うところのページ（縦８ビット、横６４ビットLCD上エリア）
	int		x_start, x_last, width;
	uint8_t	halfflg;

	if (itemsperpage == 10 || itemsperpage == 5){
		page = no2page_10[no];
		if (no >= 5){	// LCDの右半分エリアに出力
			width = LCD_WIDTH / 2 - 1;
			x_start = page *LCD_WIDTH + LCD_WIDTH / 2 + 1;
			x_last = page *LCD_WIDTH + LCD_WIDTH;
		} else {		// LCDの左半分エリアに出力するか、左から右まで使って出力する
			x_start = page *LCD_WIDTH;
			if (itemsperpage == 10){	// LCDの左半分エリアに出力する
				x_last = page *LCD_WIDTH + LCD_WIDTH / 2 - 1;
				width = LCD_WIDTH / 2 - 1;
			} else {	// 左から右まで使って出力する
				x_last = page *LCD_WIDTH + LCD_WIDTH;
				width = LCD_WIDTH;
			}
		}
		halfflg = halfflg_10[no];
	} else {	// 8 or 4
		page = no2page_8[no];
		if (no >= 4){	// LCDの右半分エリアに出力
			x_start = page *LCD_WIDTH + LCD_WIDTH / 2 + 1;
			x_last = page *LCD_WIDTH + LCD_WIDTH;
			width = LCD_WIDTH / 2 - 1;
		} else {		// LCDの左半分エリアに出力するか、左から右まで使って出力する
			x_start = page *LCD_WIDTH;
			if (itemsperpage == 8){	// LCDの左半分エリアに出力する
				x_last = page *LCD_WIDTH + LCD_WIDTH / 2 - 1;
				width = LCD_WIDTH / 2 - 1;
			} else {	// 左から右まで使って出力する
				x_last = page *LCD_WIDTH + LCD_WIDTH;
				width = LCD_WIDTH;
			}
		}
		halfflg = halfflg_8[no];
	}

	if (halfflg == 0){
		for (x=0; x<width; x++){
			LcdImageBuff[x_start + x] = RecDirList[x];
			LcdImageBuff[x_start + LCD_WIDTH + x] = (RecDirList[RECLEN_DIRNAMEIMG_THIRD + x] & 0x0F);
		}
	} else {
		for (x=0; x<width; x++){
			LcdImageBuff[x_start + x] |= ((RecDirList[x] & 0x0F) << 4);
			LcdImageBuff[x_start + LCD_WIDTH + x] = ((RecDirList[x] & 0xF0) >> 4) | (((RecDirList[RECLEN_DIRNAMEIMG_THIRD + x] & 0x0F) << 4));
		}
	}
}


/*
 * インデックスファイル/MP3リストファイル のレコードからインデックス選択メニューイメージを作成する。
 *  RecDirList : インデックスファイルのレコード
 * 	no : メニュー項目の序数 (0～、左上から下へ0,1,2,3,... )
 */
void SetRecDirList2IndexMenuImage(uint8_t *RecDirList, uint8_t no){
	static uint8_t	no2page[4]   = {0, 1, 3, 4};			// インデックスの１～４行目の各行の表示開始ページ
	static uint8_t	halfflg[4]   = {0, 1, 0, 1};			// 0:1ページ8ドット+(1/2)ページ4ドットで描画、1:(1/2)ページ４ドット+１ページ8ドットで描画

	int		x, page;
	int		x_start;

	page = no2page[no % 4];

	x_start = (no / 4) * IL_WIDTH_SELINDEX_ITEM;

	if (halfflg[no % 4] == 0){
		for (x=0; x<IL_WIDTH_SELINDEX_ITEM; x++){
			LcdImageBuff[page * LCD_WIDTH + x_start + x] = RecDirList[x];
			LcdImageBuff[(page + 1) * LCD_WIDTH + x_start + x] = (RecDirList[RECLEN_DIRNAMEIMG_THIRD + x] & 0x0F);
		}
	} else {
		for (x=0; x<IL_WIDTH_SELINDEX_ITEM; x++){
			LcdImageBuff[page * LCD_WIDTH + x_start + x] |= ((RecDirList[x] & 0x0F) << 4);
			LcdImageBuff[(page + 1) * LCD_WIDTH + x_start + x] = (((RecDirList[x] & 0xF0) >> 4) | ((RecDirList[RECLEN_DIRNAMEIMG_THIRD + x] & 0x0F) << 4));
		}
	}
}

/*
 * メニュー項目の選択状態を反転する。１２DOTフォントを前提。総項目数itemsperpageは１０または５
 * 	１０のときのメニュー項目番号は
 * 		0	5
 * 		1	6
 * 		2	7
 * 		3	8
 * 		4	9
 * 	４０のときも同様に左上から下へ0,1,2,3,...の並び。
 */
void SetMenuItemSelected(uint8_t selecteditem, uint8_t itemsperpage){
	int line;
	int x1;
	int x2;

	switch (itemsperpage){
	case 10:	// 12dotフォントで5行×２列
		line = selecteditem % 5;
		if (selecteditem < 5){
			x1 = 0;
			x2 = LCD_WIDTH/2 - 1;
		} else {
			x1 = LCD_WIDTH/2;
			x2 = LCD_WIDTH - 1;
		}
		break;
	case 5:		// 12dotフォントで５行×１列
		line = selecteditem;
		x1 = 0;
		x2 = LCD_WIDTH - 1;
		break;
	case NUM_SELINDEX_ITEM:	// インデックス選択ページ
		line = selecteditem % SELINDEX_LINES;
		x1 = (selecteditem / SELINDEX_LINES) * IL_WIDTH_SELINDEX_ITEM;
		x2 = ((selecteditem / SELINDEX_LINES) +  1) * IL_WIDTH_SELINDEX_ITEM - 1;
		break;
	default:
		return;
	}

	invert_lcd_disp_data(LcdImageBuff, DL_DOTLINEHEIGHT, line, x1, x2);
}

// LCD全体をフル表示
void DispLcdFullImage(){
	//onBoardLed2_On();
	display_lcd_data(LcdImageBuff);
	//onBoardLed2_Off();
}

/*
 * 表示項目が使うGCLDのページ番号を取得。
 * 12dotフォントを前提しているので、取得するページとその次のページが使われることになる。
 */
uint8_t GetUsePage(uint8_t item, uint8_t itemsperpage){
	static uint8_t	line2page_10[5] = {0, 1, 3, 4, 6};	//
	int line;

	switch (itemsperpage){
	case 10:	// 12dotフォントで5行×２列
		line = item % 5;
		break;
	case 5:		// 12dotフォントで５行×１列
		line = item;
		break;
	case NUM_SELINDEX_ITEM:	// インデックス選択ページ
		line = item % SELINDEX_LINES;
		break;
	default:
		return 0xFF;
	}

	return line2page_10[line];
}

/*
 * 表示項目が使うGCLDのページのX軸報告の座標範囲を取得。
 * x[0]-x[1]:CS1の使用領域、x[2]-x[3]:CS2の使用領域
 */
void GetUseXaxis(uint8_t x[], uint8_t item, uint8_t itemsperpage){
	int	col;
	uint8_t	xL, xR;

	switch (itemsperpage){
	case 10:	// 12dotフォントで5行×２列
		if (item < 5){		// CS1だけ使う
			x[0] = 0;
			x[1] = 63;
			x[2] = x[3] = 0xFF;
		} else {
			x[2] = 0;
			x[3] = 63;
			x[0] = x[1] = 0xFF;
		}
		break;
	case 5:		// 12dotフォントで５行×１列
		x[0] = x[2] = 0;
		x[1] = x[3] = 63;
		break;
	case NUM_SELINDEX_ITEM:	// インデックス選択ページ
		col = item / SELINDEX_LINES;
		xL = col * IL_WIDTH_SELINDEX_ITEM;
		xR = (col + 1) * IL_WIDTH_SELINDEX_ITEM - 1;
		if (xL < 64 && xR < 64){
			x[0] = xL;
			x[1] = xR;
			x[2] = x[3] = 0xFF;
		} else if (xL >= 64 && xR >= 64){
			x[2] = xL - 64;
			x[3] = xR - 64;
			x[0] = x[1] = 0xFF;
		} else if (xL < 64 && xR >= 64){
			x[0] = xL;
			x[1] = 63;
			x[2] = 0;
			x[3] = xR - 64;
		}
		break;
	default:
		break;
	}
}

/*
 * 上下に並んだメニュー項目の再表示。
 */
void DispMenuItemsUpDown(uint8_t item1, uint8_t item2, uint8_t itemsperpage){
	uint8_t	page1, page2, numpages;
	uint8_t pages[4];
	uint8_t	x[4];
	uint8_t	p;

	// page1が上、page2が下
	if (item1 < item2){
		page1 = GetUsePage(item1, itemsperpage);
		page2 = GetUsePage(item2, itemsperpage);
	} else {
		page1 = GetUsePage(item2, itemsperpage);
		page2 = GetUsePage(item1, itemsperpage);
	}

	pages[0] = page1;
	pages[1] = page1 + 1;
	if (page2 == pages[1]){
		pages[2] = page2 + 1;
		numpages = 3;
	} else {
		pages[2] = page2;
		pages[3] = page2 + 1;
		numpages = 4;
	}

	// item1,item2はX軸範囲が共通
	GetUseXaxis(x, item1, itemsperpage);

	if (x[0] != 0xFF){
		for (p=0; p<numpages; p++){
			display_lcd_cs1(LcdImageBuff, pages[p], x[0], x[1]);
		}
	}

	if (x[2] != 0xFF){
		for (p=0; p<numpages; p++){
			display_lcd_cs2(LcdImageBuff, pages[p], x[2], x[3]);
		}
	}

}

/*
 * 左右に並んだメニュー項目の再表示。
 */
void DispMenuItemsLeftRight(uint8_t item1, uint8_t item2, uint8_t itemsperpage){
	uint8_t	page1, page;
	uint8_t	x1[4];
	uint8_t	x2[4];

	// item1,item2は同じページを使う
	page1 = GetUsePage(item1, itemsperpage);

	if (item1 < item2){
		GetUseXaxis(x1, item1, itemsperpage);
		GetUseXaxis(x2, item2, itemsperpage);
	} else {
		GetUseXaxis(x1, item2, itemsperpage);
		GetUseXaxis(x2, item1, itemsperpage);
	}

	for (page=page1; page<=(page1 + 1); page++){
		if (x1[0] != 0xFF){
			display_lcd_cs1(LcdImageBuff, page, x1[0], x1[1]);
		}
		if (x2[0] != 0xFF){
			display_lcd_cs1(LcdImageBuff, page, x2[0], x2[1]);
		}
		if (x1[2] != 0xFF){
			display_lcd_cs2(LcdImageBuff, page, x1[2], x1[3]);
		}
		if (x2[2] != 0xFF){
			display_lcd_cs2(LcdImageBuff, page, x2[2], x2[3]);
		}
	}
}


/*
 * ランダム再生
 * 	再生中のボタン機能
 * 	NEXTボタン：同一フォルダの次のトラックへ
 * 	PLAYボタン：ランダム再生の次のディレクトリへ
 * 	DESELECTボタン：メインメニューへ
 */
int PlayRandom(uint8_t *RecDirList, FIL *Fil){
    FRESULT     res;
    int        	ret;
    uint16_t    numdirs;
    uint16_t    numsnds;
    uint16_t    dirno;
    uint16_t    start_trackno;
    uint32_t    ofs;
    uint16_t    lastdirno;

    ret = FR_OK;

    // random play start
    res = f_open(Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        return res;
    }
	//DebugWriteln((char *)"DIRLISTFILE open \r\n");
    // header record
    res = f_read_DIRLISTFILE(Fil, RecDirList);
    if (res != FR_OK) return res;
    //numdirs = *(uint16_t *)(&RecDirList[0]);
    memcpy(&numdirs, RecDirList, 2);
	//DebugWritelnV1((char *)"read header record / numdirs", numdirs);
	AutoPoffTimer = 0;

    while (1){
        if (PonRestart == 1){
            memcpy(&lastdirno, PlayDirHist + HS_LASTDIR, 2);
            if (lastdirno >= 0 && lastdirno < numdirs){
                dirno = lastdirno;
                //DebugWritelnV1("PonRestart dirno", dirno);
                memcpy(&start_trackno, PlayDirHist + HS_LASTTRACK, 2);
            } else {
                PonRestart = 0;
                continue;
            }
        } else {
            memcpy(&dirno, PlayDirHist + HS_NEXTRANDOMDIR, 2);
            if (!(dirno >= 0 && dirno < numdirs)) dirno = 0;
            start_trackno = 0;
        }
    	//DebugWritelnV1((char *)"dirno", dirno);
        ofs = (uint32_t)(dirno + 1) * (uint32_t)RECLEN_DIRLIST_FULL;     // +1 = header record
        res = f_lseek(Fil, ofs);
        if (res == FR_OK){
            res = f_read_DIRLISTFILE(Fil, RecDirList);
            if (res == FR_OK){
                numsnds = *(RecDirList + DL_NUM_SNDFILE);    // number of sound files
                if (numsnds > 0){
                    if (!(start_trackno >= 0 && start_trackno < numsnds)){
                        start_trackno = 0;
                    }
                    if (PonRestart == 0){
                        memcpy(PlayDirHist + HS_NEXTRANDOMDIR, RecDirList + DL_RECNO_NEXTRANDOMDIR, 2);
                    }
                    //
                	//DebugWriteln((char *)"PlayOrSelectMp3InSubdirectory \r\n");
                    ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno, numsnds, start_trackno, PS_PLAY);
                    if (ret == POWEROFF){
                        break;
                    } else if (ret == GOTO_MAINMENU){
                        break;
                    }
                } else {
                	//DebugWritelnV1((char *)"numsnds", numsnds);
                }
            } else {
            	//DebugWritelnV1((char *)"f_read_DIRLISTFILE", res);
            }
        } else {
        	//DebugWritelnV1((char *)"f_lseek", res);
        }
        PonRestart = 0;
    }

    f_close(Fil);

	//DebugWriteln((char *)"PlayRandom end \r\n");
    return (ret);
}

/*
 * MP3再生
 * 	mp3file:MP3ファイルフルパス、dirno:ルートディレクトリ=0から始まる行番号
 */
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno){
    FRESULT     res;
    int         ret;

    ret = FR_OK;
	//DebugWriteln((char *)(mp3file));
	res = f_open(pfilMP3, mp3file, FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		return res;
	}
	//DebugWriteln((char *)"f_open OK");

	if (PonRestart == 1){       // POWER ON RESTART
		DWORD	lastbyte;
		memcpy(&lastbyte, PlayDirHist + HS_LASTBYTE, 4);
		PonRestart = 0;
		if (f_lseek(pfilMP3, lastbyte) != FR_OK){      // go to restart point
			f_lseek(pfilMP3, 0);
		}
	}
	//f_lseek(pfilMP3, 4*32*100000L);		// 途中から再生テスト

    memcpy(PlayDirHist + HS_LASTDIR, &dirno, 2);
    *(uint16_t *)(PlayDirHist + HS_LASTTRACK) = trackno;
    *(uint32_t *)(PlayDirHist + HS_LASTBYTE) = f_tell(pfilMP3);

	i2s_init();		// i2s.c
	dac_s = dac_e = dacbuf;
	dac_s2 = dac_e2 = dacbuf;
	buftab[0] = dacbuf;
	buftab[1] = &dacbuf[1152*2];
	dacflg = 0;

	NowPlaying = 1;
	ret = decode();
	NowPlaying = 0;

    *(uint32_t *)(PlayDirHist + HS_LASTBYTE) = f_tell(pfilMP3);
	f_close(pfilMP3);

	//if (ret == GOTO_MAINMENU || ret == SELECT_OTHER_MP3){	// 中断のときだけ履歴保存
		SaveHistory(pfilMP3);
	//}

	return ret;
}

/*
 * 再生中の表示イメージ
 */
void SetImageAtPlaying(FIL *Fil, uint8_t *RecMp3List, uint8_t trackno, uint16_t numsndfile, uint8_t NumSelItem){
	uint8_t		start_trackno;
	uint8_t		t;
    UINT        br;

	if ((trackno - 2) >= 0){
		start_trackno = trackno - 2;	// 再生中の曲が中央になるように
	} else {
		start_trackno = 0;
	}

	memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
    f_lseek(Fil, (uint32_t)(start_trackno)*(RECLEN_MP3LIST_FULL));
	for (t=start_trackno; (t - start_trackno)<NumSelItem && t<numsndfile; t++){
        if (f_read(Fil, RecMp3List, RECLEN_MP3LIST_FULL, &br) != FR_OK || br != RECLEN_MP3LIST_FULL) break;
		SetRecDirList2MenuImage(RecMp3List + RECLEN_MP3LIST, t - start_trackno, NumSelItem);
	}
	SetMenuItemSelected(trackno - start_trackno, NumSelItem);
}


/*
 * 履歴ファイルを読み込む
 */
int ReadHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    res = f_open(Fil, HISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        f_read(Fil, PlayDirHist, RECLEN_HISTFILE, &br);
        f_close(Fil);
        //DebugWritelnV1("ReadHistory 0", PlayDirHist[0]);
        //DebugWritelnV1("ReadHistory 1", PlayDirHist[1]);

        return TRUE;
    }
    return FALSE;
}

/*
 * 履歴ファイルを保存
 */
void SaveHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    res = f_open(Fil, HISTFILE, FA_WRITE | FA_OPEN_EXISTING);
    if (res == FR_OK){
        //DebugWritelnV1("SaveHistory 0", PlayDirHist[0]);
        //DebugWritelnV1("SaveHistory 1", PlayDirHist[1]);

        res = f_write(Fil, PlayDirHist, RECLEN_HISTFILE, &br);
        f_close(Fil);
    }
}

// Speaker オン／オフ切り替え
void ToggleSpeakerOnOff(){
	if (SpeakerOn == 0){
		LPC_GPIO0->FIOSET |= (0x01 << SPEAKER_MUTE_BIT);	// P0.26 (set high)
		SpeakerOn = 1;
	} else {
		LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_MUTE_BIT);	// P0.26 (set low)
		SpeakerOn = 0;
	}
}

// STEREO/MONO 切り替え
void ToggleChannelModeOnOff(){
	ChannelMode = 1 - ChannelMode;
}

// 以下、デバッグ用

void onBoardLed1_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 18);
}

void onBoardLed1_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 18);
}

void onBoardLed2_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 20);
}

void onBoardLed2_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 20);
}

void onBoardLed3_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 21);
}

void onBoardLed3_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 21);
}

void onBoardLed4_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 23);
}

void onBoardLed4_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 23);
}

#if 0
void DebugWriteln(char *buff){
	UARTSendCr(1, buff, strlen(buff));
}

void DebugWritelnV1(char *msg, int val){
	char	buff[64], buff2[16];
	strcpy(buff, msg);
	itoa(val, buff2, 16);
	strcat(buff, ":");
	strcat(buff, buff2);
	strcat(buff, "\r\n");
	DebugWriteln(buff);
}
#endif

